import math
import shutil
from langumo.building import Builder
from langumo.utils import AuxiliaryFile, AuxiliaryFileManager, colorful
from typing import Tuple


class SplitValidation(Builder):
    def __init__(self, val_ratio: float = 0.1):
        self.val_ratio = val_ratio

    def _total_lines_in_file(self, af: AuxiliaryFile) -> int:
        total_lines = 0
        with af.open('rb') as fp:
            for _ in fp:
                total_lines += 1
        return total_lines

    def build(self, afm: AuxiliaryFileManager, corpus: AuxiliaryFile
              ) -> Tuple[AuxiliaryFile, AuxiliaryFile]:
        train_dataset = afm.create()
        val_dataset = afm.create()

        total_lines = self._total_lines_in_file(corpus)
        print(colorful.render(f'<r>[*]</r> split validation corpus - '
                              f'<m>{math.ceil(total_lines * self.val_ratio)} '
                              f'</m> of <m>{total_lines}</m> lines'))

        with corpus.open('rb') as src, train_dataset.open('wb') as tdst, \
                val_dataset.open('wb') as vdst:
            # Write the first `val_ratio` lines to the validation dataset file.
            for i, line in enumerate(src):
                vdst.write(line)
                if i + 1 >= total_lines * self.val_ratio:
                    break

            # After writing the validation dataset, copy the entire rest lines
            # to the train dataset.
            shutil.copyfileobj(src, tdst)

        return train_dataset, val_dataset



import os
import tqdm
from tokenizers import Tokenizer
from tokenizers.models import WordPiece
from tokenizers.trainers import WordPieceTrainer
from tokenizers.normalizers import BertNormalizer
from tokenizers.pre_tokenizers import BertPreTokenizer
from tokenizers.decoders import WordPiece as WordPieceDecoder
from langumo.building import Builder
from langumo.utils import AuxiliaryFile, AuxiliaryFileManager, colorful
from typing import List

# Enable multi-processing in tokenization.
os.environ['TOKENIZERS_PARALLELISM'] = 'true'


class TrainTokenizer(Builder):
    def __init__(self,
                 vocab_size: int = 32000,
                 subset_size: int = 512000000,
                 limit_alphabet: int = 6000,
                 unk_token: str = '[UNK]',
                 special_tokens: List[str] = []):
        self.vocab_size = vocab_size
        self.subset_size = subset_size
        self.limit_alphabet = limit_alphabet
        self.unk_token = unk_token
        self.special_tokens = special_tokens

    def _create_subset_file(self, afm: AuxiliaryFileManager, af: AuxiliaryFile
                            ) -> AuxiliaryFile:
        subset = afm.create()
        with af.open('rb') as src, subset.open('wb') as dst:
            while True:
                line = src.readline()
                if not line:
                    break

                dst.write(line)

                # If total amount of copied data is more than `subset_size`
                # then stop copying data to the subset file.
                if src.tell() > self.subset_size:
                    break
        return subset

    def build(self, afm: AuxiliaryFileManager, corpus: AuxiliaryFile
              ) -> AuxiliaryFile:
        subset = self._create_subset_file(afm, corpus)

        # Create WordPiece model with a normalizer and pre-tokenizer. Note that
        # BERT-specific normalizer and pre-tokenizer are used in this model.
        tokenizer = Tokenizer(WordPiece())
        tokenizer.normalizer = BertNormalizer(strip_accents=False)
        tokenizer.pre_tokenizer = BertPreTokenizer()

        # Train tokenizer model with subset of corpus.
        trainer = WordPieceTrainer(
            vocab_size=self.vocab_size,
            min_frequency=2,
            show_progress=True,
            limit_alphabet=self.limit_alphabet,
            special_tokens=[self.unk_token] + self.special_tokens,
            continuing_subword_prefix='##')
        tokenizer.train(trainer, [subset.name])

        # Save trained vocabulary to an auxiliary output file.
        vocab = afm.create()
        tokenizer.model.save(os.path.dirname(vocab.name))

        os.rename(os.path.join(os.path.dirname(vocab.name), 'vocab.txt'),
                  vocab.name)

        return vocab


class TokenizeSentences(Builder):
    def __init__(self,
                 unk_token: str,
                 special_tokens: List[str] = [],
                 batch_size: int = 10000):
        self.unk_token = unk_token
        self.special_tokens = special_tokens
        self.batch_size = batch_size

    def _total_lines_in_file(self, af: AuxiliaryFile) -> int:
        total_lines = 0
        with af.open('rb') as fp:
            for _ in fp:
                total_lines += 1
        return total_lines

    def build(self,
              afm: AuxiliaryFileManager,
              corpus: AuxiliaryFile,
              vocab: AuxiliaryFile) -> AuxiliaryFile:
        total_lines = self._total_lines_in_file(corpus)

        # Create WordPiece model and add special tokens. Note that `unk_token`
        # is also a special token.
        tokenizer = Tokenizer(WordPiece(vocab.name, unk_token=self.unk_token))
        tokenizer.add_special_tokens(self.special_tokens + [self.unk_token])

        # Use BERT-specific normalizer, pre-tokenizer and decoder.
        tokenizer.normalizer = BertNormalizer(strip_accents=False)
        tokenizer.pre_tokenizer = BertPreTokenizer()
        tokenizer.decoder = WordPieceDecoder(prefix='##')

        tokenized = afm.create()
        with corpus.open('r') as src, tokenized.open('w') as dst:
            # Create tqdm progress bar with colorful description.
            tqdm_iter = tqdm.tqdm(
                src,
                desc=colorful.render('<r>[*]</r> tokenize sentences with '
                                     '<g>WordPiece</g> model'),
                total=total_lines)

            batch_lines = []
            for line in tqdm_iter:
                batch_lines.append(line)

                # Encode the grouped batch sentences and write the tokenized
                # sentences to the auxiliary output file.
                if len(batch_lines) > self.batch_size:
                    for t in tokenizer.encode_batch(batch_lines):
                        dst.write(' '.join(t.tokens) + '\n')
                    batch_lines.clear()

            # Encode the remainders and write to the output file.
            if batch_lines:
                for t in tokenizer.encode_batch(batch_lines):
                    dst.write(' '.join(t.tokens) + '\n')

        return tokenized



from multiprocessing import Process, Queue
from langumo.building import Builder
from langumo.utils import (AuxiliaryFile, AuxiliaryFileManager, colorful,
                           SentenceSplitter)
from typing import Iterable


class Parser:
    def prepare(self, raw: AuxiliaryFile):
        pass

    def extract(self, raw: AuxiliaryFile) -> Iterable[str]:
        raise NotImplementedError('this method must be implemented by '
                                  'inheritor.')

    def parse(self, text: str) -> str:
        raise NotImplementedError('this method must be implemented by '
                                  'inheritor.')


class ParseRawFile(Builder):
    def __init__(self,
                 parser: Parser,
                 lang: str,
                 min_len: int,
                 max_len: int,
                 newline: str = '[NEWLINE]',
                 num_workers: int = 1):
        self.parser = parser
        self.lang = lang
        self.min_len = min_len
        self.max_len = max_len
        self.newline = newline
        self.num_workers = num_workers

    def _parse_worker(self, from_queue: Queue, to_queue: Queue):
        splitter = SentenceSplitter(self.lang)

        while True:
            # Get raw-formatted document from main process.
            document = from_queue.get()
            if document is None:
                to_queue.put(None)
                break

            # Parse the document to the plain text.
            parsed = self.parser.parse(document)

            # Divide the document into sequences with required length.
            group_sentences = []
            for paragraph in parsed.splitlines():
                for sentence in splitter.tokenize(paragraph):
                    group_sentences.append(sentence)

                    if sum(len(s) for s in group_sentences) > self.max_len:
                        to_queue.put(' '.join(group_sentences))
                        group_sentences.clear()

                # Use custom line-break token instead of `\n` which is used for
                # separating sequences.
                if group_sentences:
                    group_sentences.append(self.newline)

            # Use the remainder in dataset if its length is suitable.
            if group_sentences and group_sentences[-1] == self.newline:
                group_sentences = group_sentences[:-1]

            text = ' '.join(group_sentences)
            if len(text) > self.min_len and len(text) < self.max_len:
                to_queue.put(text)

    def _collect_worker(self, parsed: AuxiliaryFile, to_queue: Queue):
        terminated = 0
        with parsed.open('w') as fp:
            while terminated < self.num_workers:
                text = to_queue.get()
                if text is None:
                    terminated += 1
                    continue

                text += '\n' if not text.endswith('\n') else ''
                fp.write(text)

    def build(self, afm: AuxiliaryFileManager, raw: AuxiliaryFile
              ) -> AuxiliaryFile:
        parsed = afm.create()
        self.parser.prepare(raw)

        # Create processes for parsing texts in parallel and a process for
        # collecting the parsed texts and saving to the auxiliary file.
        from_queue, to_queue = Queue(), Queue()
        parsers = [Process(target=self._parse_worker,
                           args=(from_queue, to_queue),
                           daemon=True)
                   for _ in range(self.num_workers)]
        collector = Process(target=self._collect_worker,
                            args=(parsed, to_queue),
                            daemon=True)

        # Start the processes.
        print(colorful.render(f'<r>[*]</r> parse raw-formatted corpus file '
                              f'with <g>{self.parser.__class__.__name__}</g>'))

        for p in parsers:
            p.start()
        collector.start()

        # Feed the extracted raw-formatted document to each parser process.
        for document in self.parser.extract(raw):
            from_queue.put(document)
        for _ in range(self.num_workers):
            from_queue.put(None)

        # Wait for terminating the processes.
        for p in parsers:
            p.join()
        collector.join()

        return parsed



from langumo.utils import AuxiliaryFile, AuxiliaryFileManager
from typing import Union, Tuple


class Builder:
    def build(self,
              afm: AuxiliaryFileManager,
              *inputs: AuxiliaryFile
              ) -> Union[None, AuxiliaryFile, Tuple[AuxiliaryFile, ...]]:
        raise NotImplementedError('this method must be implemented by '
                                  'inheritor.')


class BuildPipeline(Builder):
    def __init__(self, *builders: Builder):
        self.builders = builders

    def build(self,
              afm: AuxiliaryFileManager,
              *inputs: AuxiliaryFile
              ) -> Union[None, AuxiliaryFile, Tuple[AuxiliaryFile, ...]]:
        with afm.auxiliary_scope():
            outputs = inputs
            for builder in self.builders:
                outputs = builder.build(afm, *outputs)

                # Lock output auxiliary files to protect from deleting for
                # passing to next builder.
                if isinstance(outputs, AuxiliaryFile):
                    outputs.lock()
                    outputs = (outputs,)
                elif isinstance(outputs, tuple):
                    for af in outputs:
                        if not isinstance(af, AuxiliaryFile):
                            raise TypeError(f'element {type(af)} is not an '
                                            f'auxiliary file.')
                        af.lock()
                elif outputs is None:
                    outputs = tuple()
                else:
                    # If the output of builder is not one of the allowed types
                    # (auxiliary file, tuple of auxiliary files, and None) then
                    # throw exception.
                    raise TypeError(f'output type {type(outputs)} from '
                                    f'builder is not allowed.')

                # Delete all unnecessary files except inputs and locked files.
                afm.clear()
        return outputs

    def run(self, parent: str):
        with AuxiliaryFileManager(parent) as afm:
            self.build(afm)

            # After running the build pipeline, delete all created dummy files
            # even though the remainders would be removed in `__exit__` of the
            # manager.
            afm.clear()



import re
import colorama
from typing import List, Dict

# Initialize `colorama` to convert formatted texts for Windows console
# automatically.
colorama.init()

# The below regex patterns are for parsing color tags in formatted string.
_re_tag_simple = re.compile(r'<(.*?)>.*?</\1>')
_re_tag_detail = re.compile(r'<(.*?)>(.*?)</\1>')


def _parse_format_tags(fstring: str) -> List[Dict[str, str]]:
    normal = [{'text': text} for text in _re_tag_simple.split(fstring)[::2]]
    colorful = [{'color': tag.group(1), 'text': tag.group(2)}
                for tag in _re_tag_detail.finditer(fstring)]

    # Return the parsed elements with preserving their orders.
    return list(sum(zip(normal, colorful), ())) + normal[-1:]


def _encode_style_element(elem: Dict[str, str]) -> str:
    prefix = ''
    if 'color' in elem:
        if elem['color'] == 'k':
            prefix = colorama.Fore.BLACK
        elif elem['color'] == 'r':
            prefix = colorama.Fore.RED
        elif elem['color'] == 'g':
            prefix = colorama.Fore.GREEN
        elif elem['color'] == 'y':
            prefix = colorama.Fore.YELLOW
        elif elem['color'] == 'b':
            prefix = colorama.Fore.BLUE
        elif elem['color'] == 'm':
            prefix = colorama.Fore.MAGENTA
        elif elem['color'] == 'c':
            prefix = colorama.Fore.CYAN
        elif elem['color'] == 'w':
            prefix = colorama.Fore.WHITE
        else:
            raise ValueError(f'color [{elem["color"]}] is not supported.')

    return prefix + elem['text'] + colorama.Style.RESET_ALL


def render(fstring: str) -> str:
    return ''.join(_encode_style_element(elem)
                   for elem in _parse_format_tags(fstring))



import argparse
import importlib
from langumo.configuration import BuildConfig
from langumo.building import (BuildPipeline, ImportFrom, ExportTo,
                              StackOutputs, Parser, ParseRawFile, MergeFiles,
                              ShuffleLines, Residual, TrainTokenizer,
                              TokenizeSentences, SplitValidation)


def _create_parser(name: str) -> Parser:
    module, cls = name[:name.rfind('.')], name[name.rfind('.') + 1:]
    return getattr(importlib.import_module(module), cls)()


def _main():
    parser = argparse.ArgumentParser(
        prog='langumo',
        description='The unified corpus building environment for Language '
                    'Models.')
    parser.add_argument('config', nargs='?', default='build.yml',
                        help='langumo build configuration')
    args = parser.parse_args()

    cfg = BuildConfig(args.config)
    BuildPipeline(
        StackOutputs((
            ImportFrom(arg['path']),
            ParseRawFile(_create_parser(arg['parser']),
                         lang=cfg['build.parsing.language'],
                         min_len=cfg['build.parsing.min-length'],
                         max_len=cfg['build.parsing.max-length'],
                         newline=cfg['build.parsing.newline-token'],
                         num_workers=cfg['build.parsing.num-workers'])
        ) for arg in cfg['inputs']),
        MergeFiles(),

        ShuffleLines(best_seek_cnt=100000, max_buckets=1024),

        Residual(
            ImportFrom(cfg['build.tokenization.prebuilt-vocab'])
            if 'build.tokenization.prebuilt-vocab' in cfg
            else TrainTokenizer(
                vocab_size=cfg['build.tokenization.vocab-size'],
                subset_size=cfg['build.tokenization.subset-size'],
                limit_alphabet=cfg['build.tokenization.limit-alphabet'],
                unk_token=cfg['build.tokenization.unk-token'],
                special_tokens=cfg['build.tokenization.special-tokens']),

            ExportTo(cfg['outputs.vocabulary'])
        ),

        TokenizeSentences(
            unk_token=cfg['build.tokenization.unk-token'],
            special_tokens=cfg['build.tokenization.special-tokens'],
            batch_size=10000),
        SplitValidation(val_ratio=cfg['build.splitting.validation-ratio']),
        ExportTo(cfg['outputs.train-corpus'], cfg['outputs.eval-corpus'])
    ).run(cfg['workspace'])



from langumo.building.base import Builder, BuildPipeline
from langumo.building.miscellaneous import (ImportFrom, ExportTo, Residual,
                                            StackOutputs)
from langumo.building.parsing import Parser, ParseRawFile
from langumo.building.shuffling import ShuffleLines
from langumo.building.tokenization import TrainTokenizer, TokenizeSentences
from langumo.building.splitting import SplitValidation
from langumo.building.mergence import MergeFiles



import re
import bz2
import mwparserfromhell as mw
import xml.etree.cElementTree as etree
from langumo.building import Parser
from langumo.utils import AuxiliaryFile
from typing import Iterable


class WikipediaParser(Parser):
    brackets_pattern = re.compile(r'\([^(]*?\)')
    single_quotes_pattern = re.compile('[\x60\xb4\u2018\u2019]')
    double_quotes_pattern = re.compile('[\u201c\u201d]')

    def __init__(self):
        self.namespaces = []

    def prepare(self, raw: AuxiliaryFile):
        with bz2.open(raw.name, 'r') as fp:
            context = etree.iterparse(fp, events=('start', 'end'))

            # Collect namespaces from the dump file.
            for event, elem in context:
                if event != 'end':
                    continue

                if elem.tag.endswith('namespace'):
                    if elem.text is not None:
                        self.namespaces.append(elem.text)

                # Exit the loop if the parser meets the end of `<namespaces>`
                # tag.
                if elem.tag.endswith('namespaces'):
                    break

    def extract(self, raw: AuxiliaryFile) -> Iterable[str]:
        with bz2.open(raw.name, 'r') as fp:
            context = etree.iterparse(fp, events=('start', 'end'))
            _, root = next(context)

            # Parse articles from the dump file and yield them.
            prefix_pattern = re.compile(r'({.*?})')
            for event, elem in context:
                if event != 'end' or not elem.tag.endswith('page'):
                    continue

                # Skip the article which does not have namespace of 0.
                prefix = prefix_pattern.match(elem.tag).group(0)
                if elem.find(f'./{prefix}ns').text != '0':
                    root.clear()
                    continue

                # Get a mediawiki code of the article content.
                article = elem.find(f'./{prefix}revision/{prefix}text').text
                root.clear()

                if article is None or article.lower().startswith('#redirect'):
                    continue

                yield article

    def _clean_mediawiki_text(self, text: str) -> str:
        # Parse mediawiki code by using `mwparserfromhell` and create
        # namespace-based wiki-link pattern.
        wiki = mw.parse(text)
        mw_nslinks_pattern = re.compile(
            '^(?:{}):'.format('|'.join(self.namespaces)), re.IGNORECASE)

        # Simple remove wrapper function.
        def remove_element(section, obj):
            try:
                section.remove(obj)
            except ValueError:
                pass

        # Filter functions to remove from mediawiki code.
        def filter_wikilinks(obj):
            return bool(mw_nslinks_pattern.match(str(obj.title)))

        def filter_templates(obj):
            return obj.name.lower() in {'reflist', 'notelist', 'notelist-ua',
                                        'notelist-lr', 'notelist-ur',
                                        'notelist-lg'}

        def filter_tags(obj):
            return str(obj.tag) in {'ref', 'table'}

        section_text = []
        for section in wiki.get_sections(flat=True, include_lead=True):
            # Remove elements filtered by above functions.
            for obj in section.ifilter_wikilinks(matches=filter_wikilinks,
                                                 recursive=True):
                remove_element(section, obj)
            for obj in section.ifilter_templates(matches=filter_templates,
                                                 recursive=True):
                remove_element(section, obj)
            for obj in section.ifilter_tags(matches=filter_tags,
                                            recursive=True):
                remove_element(section, obj)

            # Add cleaned mediawiki-based contents.
            section_text.append(section.strip_code().strip())

        return section_text

    def parse(self, text: str) -> str:
        section_text = self._clean_mediawiki_text(text)

        # Post-process parsed wikipedia article contents.
        filtered = []
        for text in section_text:
            for line in text.strip().splitlines():
                # Check if the text has normal punctuation.
                if not line or line[-1] not in '!?.':
                    continue

                # Remove nested brackets and unnecessary spaces.
                while WikipediaParser.brackets_pattern.search(line):
                    line = WikipediaParser.brackets_pattern.sub('', line)

                line = line.replace('\n', ' ').replace('\t', ' ')
                while '  ' in line:
                    line = line.replace('  ', ' ')

                # Normalize the quotes by replacing unusual ones to the
                # standard ones.
                line = WikipediaParser.single_quotes_pattern.sub('\'', line)
                line = WikipediaParser.double_quotes_pattern.sub('"', line)

                filtered.append(line)

        return '\n'.join(filtered)



from langumo.utils import colorful
from typing import List


class SentenceSplitter:
    def __init__(self, lang: str):
        self.lang = lang
        self._splitter = None

        # Prepare prerequisite resources for sentence tokenizers.
        if lang == 'en':
            import nltk

            try:
                nltk.data.find('tokenizers/punkt')
            except LookupError:
                print(colorful.render('<r>[*]</r> prepare resources for '
                                      '<b>splitting sentences</b>'))

                nltk.download('punkt')

    def _initialize(self):
        if self.lang == 'en':
            import nltk

            # NLTK tokenizer needs pre-downloaded resources.
            nltk.data.find('tokenizers/punkt')
            self._splitter = nltk.tokenize.sent_tokenize
        elif self.lang == 'ko':
            import kss
            self._splitter = kss.split_sentences

    def tokenize(self, text: str) -> List[str]:
        if self._splitter is None:
            self._initialize()

        return list(self._splitter(text))



import os
import shutil
from langumo.building import Builder, BuildPipeline
from langumo.utils import AuxiliaryFile, AuxiliaryFileManager, colorful
from typing import Union, Tuple, Iterable


class ImportFrom(Builder):
    def __init__(self, *paths: str):
        self.paths = paths

    def build(self, afm: AuxiliaryFileManager, *inputs: AuxiliaryFile
              ) -> AuxiliaryFile:
        # Note that an imported files would be wrapped with `AuxiliaryFile`
        # directly. Because the files are not created by `AuxiliaryFileManager`
        # but brought simply from existing external files, they do not need to
        # be removed. Namely, the manager does not have the ownership of them.
        files = []
        for path in self.paths:
            print(colorful.render(f'<r>[*]</r> import file from '
                                  f'<b>{path}</b>'))
            files.append(AuxiliaryFile(path))
        return tuple(files)


class ExportTo(Builder):
    def __init__(self, *paths: str):
        self.paths = paths

    def build(self, afm: AuxiliaryFileManager, *inputs: AuxiliaryFile
              ) -> Union[None, AuxiliaryFile, Tuple[AuxiliaryFile, ...]]:
        if len(inputs) != len(self.paths):
            raise ValueError('number of predifined exporting files are not '
                             'matched to the given auxiliary files.')

        # Save the auxiliary files to the exporting paths.
        for af, path in zip(inputs, self.paths):
            # Create an exporting directory if not exist.
            parent = os.path.dirname(path)
            if parent != '' and not os.path.exists(parent):
                os.makedirs(parent, exist_ok=True)

            print(colorful.render(f'<r>[*]</r> export the processed file to '
                                  f'<b>{path}</b>'))
            shutil.copyfile(af.name, path)

        return inputs


class Residual(Builder):
    def __init__(self, *builders: Builder):
        self.builder = (BuildPipeline(*builders) if len(builders) > 1
                        else builders[0])

    def build(self, afm: AuxiliaryFileManager, *inputs: AuxiliaryFile
              ) -> Tuple[AuxiliaryFile, ...]:
        outputs = self.builder.build(afm, *inputs) or tuple()
        return inputs + outputs


class StackOutputs(Builder):
    def __init__(self,
                 builder_group: Iterable[Union[Builder, Tuple[Builder, ...]]]):
        self.builders = []
        for builders in builder_group:
            if isinstance(builders, tuple):
                if len(builders) == 1:
                    self.builders.append(builders[0])
                else:
                    self.builders.append(BuildPipeline(*builders))
            else:
                self.builders.append(builders)

    def build(self, afm: AuxiliaryFileManager, *inputs: AuxiliaryFile
              ) -> Tuple[AuxiliaryFile, ...]:
        outputs = tuple()
        for builder in self.builders:
            outputs += builder.build(afm, *inputs)
            afm.synchronize(outputs)

            # Lock input and stacked output auxiliary files.
            for af in inputs + outputs:
                af.lock()
            afm.clear()
        return outputs



from langumo.building import Builder
from langumo.utils import AuxiliaryFile, AuxiliaryFileManager, colorful


class MergeFiles(Builder):
    def build(self, afm: AuxiliaryFileManager, *inputs: AuxiliaryFile
              ) -> AuxiliaryFile:
        merged = afm.create()

        print(colorful.render(f'<r>[*]</r> merge <m>{len(inputs)}</m> files '
                              f'into one'))
        with merged.open('wb') as dst, \
                AuxiliaryFile.opens(inputs, 'rb') as srcs:
            for src in srcs:
                for line in src:
                    # Add break-line character to the end of text to avoid
                    # being merged with other line.
                    line += b'\n' if not line.endswith(b'\n') else b''

                    dst.write(line)

        return merged



import tqdm
import random
import shutil
from langumo.building import Builder
from langumo.utils import AuxiliaryFile, AuxiliaryFileManager, colorful
from typing import List


class ShuffleLines(Builder):
    def __init__(self, best_seek_cnt: int = 100000, max_buckets: int = 512):
        self.best_seek_cnt = best_seek_cnt
        self.max_buckets = max_buckets

    def _total_lines_in_file(self, af: AuxiliaryFile) -> int:
        total_lines = 0
        with af.open('rb') as fp:
            for _ in fp:
                total_lines += 1
        return total_lines

    def _collect_seek_offsets(self, af: AuxiliaryFile, stride: int
                              ) -> List[int]:
        offsets = []
        with af.open('rb') as fp:
            while True:
                current = fp.tell()

                # Read `stride` lines and move to the end of the chunk. If the
                # last line in the chunk is empty, then it means current is the
                # last of entire chunks.
                lines = [fp.readline() for _ in range(stride)]
                if not lines[-1]:
                    break

                # Gather the current position to the collection.
                offsets.append(current)
        return offsets

    def build(self, afm: AuxiliaryFileManager, corpus: AuxiliaryFile
              ) -> AuxiliaryFile:
        # Calculate the optimum stride and bucket size.
        stride = max(
            1, self._total_lines_in_file(corpus) // self.best_seek_cnt)
        buckets = [afm.create()
                   for _ in range(min(stride * 2, self.max_buckets))]

        # Collect the corresponding seeking positions and shuffle them.
        offsets = self._collect_seek_offsets(corpus, stride)
        random.shuffle(offsets)

        with corpus.open('rb') as src, \
                AuxiliaryFile.opens(buckets, 'wb') as dsts:
            # Create tqdm progress bar with colorful description.
            tqdm_iter = tqdm.tqdm(
                offsets,
                desc=colorful.render('<r>[*]</r> shuffle raw corpus file'))

            for offset in tqdm_iter:
                src.seek(offset)

                for _ in range(stride):
                    line = src.readline()
                    if not line:
                        break

                    # Add break-line character to the end of text to avoid
                    # being merged with other line.
                    line += b'\n' if not line.endswith(b'\n') else b''

                    # Write the decorated text line to the random bucket for
                    # ensuring randomness.
                    dsts[random.randint(0, len(dsts) - 1)].write(line)

        # After splitting to the buckets, merge them into a single file.
        merged = afm.create()
        with merged.open('wb') as dst, \
                AuxiliaryFile.opens(buckets, 'rb') as srcs:
            for src in srcs:
                shutil.copyfileobj(src, dst)

        return merged



import os
import shutil
import random
import string
import contextlib
from typing import IO, List, Container


class AuxiliaryFile:
    def __init__(self, name: str):
        self.name = name
        self.locked = False

    def lock(self):
        self.locked = True
        return self

    def open(self, mode: str = 'r') -> IO:
        return open(self.name, mode)

    @staticmethod
    @contextlib.contextmanager
    def opens(files: List['AuxiliaryFile'], mode: str = 'r') -> List[IO]:
        files = [f.open(mode) for f in files]
        try:
            yield files
        finally:
            for f in files:
                f.close()


class AuxiliaryFileManager:
    def __init__(self, parent: str):
        self.files = []
        self.parent = parent
        self.auxiliary_level = 0

        # Create parent workspace directory.
        os.makedirs(parent)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _create_random_filename(self) -> str:
        while True:
            # Create random filename which is a combination of digits and lower
            # case letters.
            name = os.path.join(self.parent, ''.join(
                random.choices(string.digits + string.ascii_lowercase, k=16)))

            if all(name != af.name for af, _ in self.files):
                return name

    def create(self) -> AuxiliaryFile:
        file = AuxiliaryFile(self._create_random_filename())
        self.files.append((file, self.auxiliary_level))

        # Create an empty file.
        with open(file.name, 'w') as fp:
            fp.close()

        return file

    def synchronize(self, files: Container[AuxiliaryFile]):
        for i, (af, level) in enumerate(self.files):
            if af in files:
                self.files[i] = (af, self.auxiliary_level)

    @contextlib.contextmanager
    def auxiliary_scope(self):
        self.auxiliary_level += 1
        try:
            yield
        finally:
            self.auxiliary_level -= 1

    def clear(self):
        for af, level in self.files.copy():
            # Skip clearing high-level files.
            if level < self.auxiliary_level:
                continue

            if af.locked:
                # If the auxiliary file is locked, then skip deleting and
                # unlock it.
                af.locked = False
            else:
                self.files.remove((af, level))
                os.remove(af.name)

    def close(self):
        # Remove the parent workspace directory and its sub-files.
        shutil.rmtree(self.parent)



import yaml
from typing import Dict, Any


_default_configuration = {
    'workspace': 'tmp',
    'outputs': {
        'vocabulary': 'build/vocab.txt',
        'train-corpus': 'build/corpus.train.txt',
        'eval-corpus': 'build/corpus.eval.txt',
    },
    'build': {
        'parsing': {
            'num-workers': 1,
            'language': 'en',
            'newline-token': '[NEWLINE]',
            'min-length': 0,
            'max-length': 1024
        },
        'splitting': {
            'validation-ratio': 0.1
        },
        'tokenization': {
            'subset-size': 1000000000,
            'vocab-size': 32000,
            'limit-alphabet': 1000,
            'unk-token': '[UNK]',
            'special-tokens': ['[START]', '[END]', '[PAD]', '[NEWLINE]']
        }
    }
}


def _update_default_dict(data: Dict, default: Dict):
    for key in default:
        if key not in data:
            data[key] = default[key]
        elif isinstance(data[key], dict) and isinstance(default[key], dict):
            _update_default_dict(data[key], default[key])


class BuildConfig:
    def __init__(self, path: str):
        with open(path, 'r') as fp:
            self.config_dict = yaml.safe_load(fp)

        if 'langumo' not in self.config_dict:
            raise RuntimeError('build configuration file should contain '
                               '`langumo` namespace.')
        self.config_dict = self.config_dict['langumo']

        # Update default configuration values.
        _update_default_dict(self.config_dict, _default_configuration)

    def __getitem__(self, key: str) -> Any:
        current = self.config_dict
        for namespace in key.split('.'):
            current = current[namespace]
        return current

    def __contains__(self, key: str) -> bool:
        current = self.config_dict
        for namespace in key.split('.'):
            if namespace not in current:
                return False
            current = current[namespace]
        return True



import re
import json
from langumo.building import Parser
from langumo.utils import AuxiliaryFile
from typing import Iterable


class EscapedJSONStringParser(Parser):
    single_quotes_pattern = re.compile('[\x60\xb4\u2018\u2019]')
    double_quotes_pattern = re.compile('[\u201c\u201d]')

    def extract(self, raw: AuxiliaryFile) -> Iterable[str]:
        with raw.open('r') as fp:
            for line in fp:
                if not line.strip():
                    continue

                yield line

    def parse(self, text: str) -> str:
        text, _ = json.decoder.scanstring(text, 1)

        filtered = []
        for line in text.strip().splitlines():
            if not line:
                continue

            # Remove duplicated spaces.
            line = line.replace('\n', ' ').replace('\t', ' ')
            while '  ' in line:
                line = line.replace('  ', ' ')

            # Normalize the quotes by replacing unusual ones to the standard
            # ones.
            line = (EscapedJSONStringParser
                    .single_quotes_pattern.sub('\'', line))
            line = (EscapedJSONStringParser
                    .double_quotes_pattern.sub('"', line))

            filtered.append(line)

        return '\n'.join(filtered)
